// -*- C++ -*-

// Top directory
// This is initialized from the cache directory, but users can
// create directories with the names of hostnames.  Directories
// are always created with the official hostname of a machine,
// with all the alises appearing as symlinks.

// (C) Copyright 1994 Jeremy Fitzhardinge <jeremy@sw.oz.au>
// This code is distributed under the terms of the
// GNU General Public Licence.  See COPYING for more details.

#pragma implementation

#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <String.h>

#include <unistd.h>
#include <netdb.h>
#include <stdio.h>

#include "ftpfs.h"
#include "topdir.h"
#include "sitetopdir.h"
#include "ftpconn.h"
#include "ftplink.h"

int 
topdir::do_create(const up_preamble &, upp_repl &,
		  const upp_create_s &arg, upp_create_r &ret)
{
	if (!S_ISDIR(arg.mode))
		return EPERM;
	
	String name((char *)arg.name.elems, (int)arg.name.nelem);
	
	return create(name, ret.file);
}
	
int
topdir::create(const String &name, up_handle &handle)
{	
	struct hostent *he = gethostbyname((const char *)name);

	if (!he)
	  {
	    cerr << "Hostname lookup for " << name << " failed:",
	    herror("");
	    return ENOENT;
	  }

	struct hostent *rhe = NULL;
	String hostname = he->h_name; // First guess

	if (!he->h_addr_list[0] || he->h_addrtype != AF_INET)
	  {
	    cerr << name << ": no address associated with name\n";
	    return ENOENT;
	  }
	    
	cout << "IP for " << name << " is "
	  << (unsigned)(unsigned char)(he->h_addr_list[0][0]) << '.'
	  << (unsigned)(unsigned char)(he->h_addr_list[0][1]) << '.'
	  << (unsigned)(unsigned char)(he->h_addr_list[0][2]) << '.'
	  << (unsigned)(unsigned char)(he->h_addr_list[0][3]) << ".\n";

	char *xaddr = new char[he->h_length];
	if (xaddr)
	  {
	    // Avoid stepping on buffers

	    memcpy(xaddr, he->h_addr_list[0], he->h_length);
	    
	    rhe = gethostbyaddr(xaddr, he->h_length, he->h_addrtype);
	    delete xaddr;
	  }
	
	cout << "Reverse lookup " << (rhe ? "succeeded\n" : "failed\n");

	if (rhe)
	  hostname = rhe->h_name; // Reverse lookup is better

	cout << "Official name of " << name << " is " << hostname << '\n';

	if (lookup(name))	// Already exists?
	  return EEXIST;
	else
	  {
	    DirEntry *de = lookup(hostname);

	    if (!de)
	      {
		Inode *ino = new sitetopdir(filesys, hostname, this);
		link(hostname, ino);
		handle.handle = ino->gethandle();
	      }
	    if (name != hostname)
	      {
		Inode *ino = new ftplink(filesys, hostname, this, name);
		if (link(name, ino))
		  delete ino;
	      }
	    if (de)
	      return EEXIST;	// We just needed a new link
	  }
	
	for (char **aliaspp = he->h_aliases; aliaspp && *aliaspp; aliaspp++)
	{
		String alias(*aliaspp);

		cout << "aliasing to " << *aliaspp << '\n';
		
		Inode *ino = new ftplink(filesys, hostname, this, alias);
		if (link(alias, ino))
			delete ino;
	}
	
	return 0;
}

topdir::topdir(ftpfs &fs, const String &cache)
       :dir(fs, cache, NULL)
{
	mode = S_IFDIR | 0777;
	init();
}

Inode *
topdir::newinode(dir *d, const struct stat *st, const String &name)
{
	Inode *ino = NULL;
	
	if (S_ISDIR(st->st_mode))
	{
		if (name == "." || name == "..")
			return NULL;
		ino = new sitetopdir(filesys, name, this);
	}
	else if (S_ISLNK(st->st_mode))
	{
		char buf[1024];
		int len = readlink(name, buf, sizeof(buf));
		if (len == -1)
			return NULL;
			
		ino = new ftplink(filesys, String(buf, len), this, name);
	}
	
	return ino;
}

//
// hpa 951118: topdir::do_lookup is overloaded in order to provide an
//             "automounter" effect -- if we are looking for a specific
//             filename, and it doesn't exist, attempt to create it
//
// Curious note: the obvious way to do this would be to re-define
// topdir::lookup (it would save a lot of duplicate work), but the
// lookup function is not virtual, so do_lookup would call
// DirInode::lookup and not topdir::lookup.  Having a non-virtual
// lookup seems to defeat the purpose, somehow...
// 
int
topdir::do_lookup(const up_preamble &pre, upp_repl &rep,
                  const upp_lookup_s &arg, upp_lookup_r &ret)
{
  String name((char *)arg.name.elems, (int)arg.name.nelem); 

  // First check if we have already come across this site...
  if ( !dir::lookup(name) )
    {
      // Not found.  Attempt to create.
      
      cout << name << " not found, attempting to connect to site.\n";
      
      up_handle bogus;		// We don't really need this information
      create(name, bogus);	// (at least not right now)

      // We don't check the result of "create", since if it failed,
      // dir::do_lookup will fail and handle all the paperwork for us. 
    }

  return dir::do_lookup(pre,rep,arg,ret);
}


